@@ -7,6 +7,7 @@ from __future__ import annotations
 
 import logging
 import os
+import pathlib
 import re
 import sys
 import uuid
@@ -40,6 +41,7 @@ from .auth import (
 from .auth.idtoken import AuthByIdToken
 from .bind_upload_agent import BindUploadError
 from .compat import IS_LINUX, IS_WINDOWS, quote, urlencode
+from .config_manager import CONFIG_PARSER
 from .connection_diagnostic import ConnectionDiagnostic
 from .constants import (
     ENV_VAR_PARTNER,
@@ -214,6 +216,14 @@ DEFAULT_CONFIGURATION: dict[str, tuple[Any, type | tuple[type, ...]]] = {
         False,
         bool,
     ),  # Whether to force the JSON content to be decoded in utf-8, it is only effective when result format is JSON
+    "server_session_keep_alive": (
+        False,
+        bool,
+    ),  # Whether to keep session alive after connector shuts down
+    "enable_retry_reason_in_query_response": (
+        True,
+        bool,
+    ),  # Enable sending retryReason in response header for query-requests
 }
 
 APPLICATION_RE = re.compile(r"[\w\d_]+")
@@ -278,11 +288,18 @@ class SnowflakeConnection:
         json_result_force_utf8_decoding: When true, json result will be decoded in utf-8,
           when false, the encoding of the content is auto-detected. Default value is false.
           This parameter is only effective when the result format is JSON.
+        server_session_keep_alive: When true, the connector does not destroy the session on the Snowflake server side
+          before the connector shuts down. Default value is false.
     """
 
     OCSP_ENV_LOCK = Lock()
 
-    def __init__(self, **kwargs) -> None:
+    def __init__(
+        self,
+        connection_name: str | None = None,
+        connections_file_path: pathlib.Path | None = None,
+        **kwargs,
+    ) -> None:
         self._lock_sequence_counter = Lock()
         self.sequence_counter = 0
         self._errorhandler = Error.default_errorhandler
@@ -315,6 +332,21 @@ class SnowflakeConnection:
         self.converter = None
         self.query_context_cache: QueryContextCache | None = None
         self.query_context_cache_size = 5
+        if connections_file_path is not None:
+            # Change config file path and force update cache
+            for i, s in enumerate(CONFIG_PARSER._slices):
+                if s.section == "connections":
+                    CONFIG_PARSER._slices[i] = s._replace(path=connections_file_path)
+                    CONFIG_PARSER.read_config()
+                    break
+        if connection_name is not None:
+            connections = CONFIG_PARSER["connections"]
+            if connection_name not in connections:
+                raise Error(
+                    f"Invalid connection_name '{connection_name}',"
+                    f" known ones are {list(connections.keys())}"
+                )
+            kwargs = {**connections[connection_name], **kwargs}
         self.__set_error_attributes()
         self.connect(**kwargs)
         self._telemetry = TelemetryClient(self._rest)
@@ -337,7 +369,8 @@ class SnowflakeConnection:
         return self._ocsp_fail_open
 
     def _ocsp_mode(self) -> OCSPMode:
-        """OCSP mode. INSECURE, FAIL_OPEN or FAIL_CLOSED."""
+        """OCSP mode. INSEC
+        URE, FAIL_OPEN or FAIL_CLOSED."""
         if self.insecure_mode:
             return OCSPMode.INSECURE
         elif self.ocsp_fail_open:
@@ -604,7 +637,10 @@ class SnowflakeConnection:
             # close telemetry first, since it needs rest to send remaining data
             logger.info("closed")
             self._telemetry.close(send_on_close=retry)
-            if self._all_async_queries_finished():
+            if (
+                self._all_async_queries_finished()
+                and not self._server_session_keep_alive
+            ):
                 logger.info("No async queries seem to be running, deleting session")
                 self.rest.delete_session(retry=retry)
             else:
@@ -1075,8 +1111,10 @@ class SnowflakeConnection:
             # binding parameters. This is for qmarks paramstyle.
             data["bindings"] = binding_params
         if not _no_results:
-            # not an async query
-            data["queryContext"] = self.get_query_context()
+            # not an async query.
+            queryContext = self.get_query_context()
+            #  Here queryContextDTO should be a dict object field, same with `parameters` field
+            data["queryContextDTO"] = queryContext
         client = "sfsql_file_transfer" if is_file_transfer else "sfsql"
 
         if logger.getEffectiveLevel() <= logging.DEBUG:
@@ -1615,10 +1653,10 @@ class SnowflakeConnection:
         if not self.is_query_context_cache_disabled:
             self.query_context_cache = QueryContextCache(self.query_context_cache_size)
 
-    def get_query_context(self) -> str | None:
+    def get_query_context(self) -> dict | None:
         if self.is_query_context_cache_disabled:
             return None
-        return self.query_context_cache.serialize_to_json()
+        return self.query_context_cache.serialize_to_dict()
 
     def set_query_context(self, data: dict) -> None:
         if not self.is_query_context_cache_disabled:
@@ -1653,10 +1691,7 @@ class SnowflakeConnection:
         if not self._async_sfqids:
             return True
 
-        if sys.version_info >= (3, 8):
-            queries = list(reversed(self._async_sfqids.keys()))
-        else:
-            queries = list(reversed(list(self._async_sfqids.keys())))
+        queries = list(reversed(self._async_sfqids.keys()))
 
         num_workers = min(self.client_prefetch_threads, len(queries))
         found_unfinished_query = False
@@ -1672,7 +1707,6 @@ class SnowflakeConnection:
         with ThreadPoolExecutor(
             max_workers=num_workers, thread_name_prefix="async_query_check_"
         ) as tpe:  # We should upgrade to using cancel_futures=True once supporting 3.9+
-
             futures = (tpe.submit(async_query_check_helper, sfqid) for sfqid in queries)
             for f in as_completed(futures):
                 if f.result():
