@@ -22,6 +22,7 @@ from typing import (
     Any,
     Callable,
     Iterator,
+    Literal,
     NamedTuple,
     NoReturn,
     Sequence,
@@ -29,7 +30,7 @@ from typing import (
     overload,
 )
 
-from typing_extensions import Literal, Self
+from typing_extensions import Self
 
 from snowflake.connector.result_batch import create_batches_from_response
 from snowflake.connector.result_set import ResultSet
@@ -264,7 +265,6 @@ class SnowflakeCursor:
         self._inner_cursor: SnowflakeCursor | None = None
         self._prefetch_hook = None
         self._rownumber: int | None = None
-        self._cached_last_result: dict | None = None
 
         self.reset()
 
@@ -543,8 +543,6 @@ class SnowflakeCursor:
                 self._timebomb.cancel()
                 logger.debug("cancelled timebomb in finally")
 
-            self._cached_last_result = ret
-
         if "data" in ret and "parameters" in ret["data"]:
             parameters = ret["data"].get("parameters", list())
             # Set session parameters for cursor object
@@ -1076,7 +1074,6 @@ class SnowflakeCursor:
             Error.errorhandler_wrapper(
                 self.connection, self, ProgrammingError, errvalue
             )
-        self._cached_last_result = ret
         return self
 
     def fetch_arrow_batches(self) -> Iterator[Table]:
@@ -1126,7 +1123,6 @@ class SnowflakeCursor:
     def abort_query(self, qid: str) -> bool:
         url = f"/queries/{qid}/abort-request"
         ret = self._connection.rest.request(url=url, method="post")
-        self._cached_last_result = ret
         return ret.get("success")
 
     def executemany(
@@ -1462,7 +1458,6 @@ class SnowflakeCursor:
             ):
                 url = f"/queries/{sfqid}/result"
                 ret = self._connection.rest.request(url=url, method="get")
-                self._cached_last_result = ret
                 if "data" in ret and "resultIds" in ret["data"]:
                     self._init_multi_statement_results(ret["data"])
 
